import os
import yaml
import json
import click
import sys
import logging
from urllib.parse import urlparse
from typing import List

from . import cli
from .params import project
from .utils import add_plugin, add_related_plugins, install_plugins
from meltano.core.plugin import PluginType
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_add_custom_service import ProjectAddCustomService
from meltano.core.plugin_install_service import PluginInstallReason


@cli.command()
@click.argument("plugin_type", type=click.Choice(PluginType.cli_arguments()))
@click.argument("plugin_name", nargs=-1, required=True)
@click.option("--custom", is_flag=True)
@click.option("--include-related", is_flag=True)
@project()
@click.pass_context
def add(ctx, project, plugin_type, plugin_name, **flags):
    plugin_type = PluginType.from_cli_argument(plugin_type)
    plugin_names = plugin_name  # nargs=-1

    if flags["custom"]:
        if plugin_type in (
            PluginType.TRANSFORMERS,
            PluginType.TRANSFORMS,
            PluginType.ORCHESTRATORS,
        ):
            click.secho(f"--custom is not supported for {ctx.invoked_subcommand}")
            raise click.Abort()

        add_service = ProjectAddCustomService(project)
    else:
        add_service = ProjectAddService(project)

    plugins = [
        add_plugin(project, plugin_type, plugin_name, add_service=add_service)
        for plugin_name in plugin_names
    ]

    related_plugin_types = [PluginType.FILES]
    if flags["include_related"]:
        related_plugin_types = list(PluginType)

    related_plugins = add_related_plugins(
        project, plugins, add_service=add_service, plugin_types=related_plugin_types
    )
    plugins.extend(related_plugins)

    success = install_plugins(project, plugins, reason=PluginInstallReason.ADD)

    for plugin in plugins:  # TODO: Only works on Plugin from discovery...
        docs_link = plugin._extras.get("docs")
        if docs_link:
            click.echo(
                f"For more details about {plugin.type.descriptor} '{plugin.name}', visit {docs_link}"
            )

    if not success:
        raise click.Abort()



import yaml
import copy


from typing import Iterable, List
from meltano.core.schedule import Schedule
from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.plugin.factory import plugin_factory
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq


VERSION = 1


class MeltanoFile(Canonical):
    def __init__(self, **attrs):
        version = int(attrs.pop("version", VERSION))
        plugins = self.load_plugins(attrs.pop("plugins", {}))
        schedules = self.load_schedules(attrs.pop("schedules", []))

        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            version=version,
            config=attrs,
            plugins=plugins,
            schedules=schedules,
        )

    def load_plugins(self, plugins) -> Canonical:
        """Parse the meltano.yml file and return it as `PluginInstall` instances."""
        plugin_type_plugins = Canonical()

        for plugin_type in PluginType:
            plugin_type_plugins[plugin_type] = []

        # this will parse the meltano.yml file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        for plugin_type, plugin_defs in plugins.items():
            for plugin_def in plugin_defs:
                plugin = plugin_factory(plugin_type, plugin_def)
                plugin_type_plugins[plugin.type].append(plugin)

        return plugin_type_plugins

    def load_schedules(self, schedules) -> List[Schedule]:
        return list(map(Schedule.parse, schedules))

    def __iter__(self):
        for k, v in super().__iter__():
            if k == "config":
                yield from v.items()
            else:
                yield (k, v)



import asyncio
import logging
import os
import re
import sys

from meltano.core.project_settings_service import ProjectSettingsService

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
DEFAULT_LEVEL = "info"
FORMAT = (
    "[%(asctime)s] [%(process)d|%(threadName)10s|%(name)s] [%(levelname)s] %(message)s"
)


def parse_log_level(log_level):
    return LEVELS.get(log_level, LEVELS[DEFAULT_LEVEL])


def setup_logging(project=None, log_level=DEFAULT_LEVEL):
    # Mimick Python 3.8's `force=True` kwarg to override any
    # existing logger handlers
    # See https://github.com/python/cpython/commit/cf67d6a934b51b1f97e72945b596477b271f70b8
    root = logging.getLogger()
    for h in root.handlers[:]:
        root.removeHandler(h)
        h.close()

    if project:
        settings_service = ProjectSettingsService(project)
        log_level = settings_service.get("cli.log_level")

    logging.basicConfig(
        stream=sys.stderr, format=FORMAT, level=parse_log_level(log_level)
    )


def remove_ansi_escape_sequences(line):
    """
    Remove ANSI escape sequences that are used for adding colors in
     terminals, so that only the text is logged in a file
    """
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", line)


async def capture_subprocess_output(stream_captured, output_stream):
    """
    Capture in real time the output stream of a suprocess that is run async.
    The stream has been set to asyncio.subprocess.PIPE and is provided using
     stream_captured to this function.
    As new lines are captured for stream_captured, they are written to output_stream.
    This async function should be run with await asyncio.wait() while waiting
     for the subprocess to end.
    """
    while not stream_captured.at_eof():
        line = await stream_captured.readline()
        if not line:
            continue

        output_stream.write(line.decode())



import os
import io
import yaml
import requests
import logging
import shutil
import re
from copy import deepcopy
from typing import Dict, Iterator, Optional
from itertools import groupby, chain

import meltano.core.bundle as bundle
from .project_settings_service import ProjectSettingsService
from .setting_definition import SettingDefinition
from .behavior.versioned import Versioned, IncompatibleVersionError
from .behavior.canonical import Canonical
from .config_service import ConfigService
from .plugin import Plugin, PluginInstall, PluginType, PluginRef
from .plugin.factory import plugin_factory


class PluginNotFoundError(Exception):
    pass


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""

    pass


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""

    pass


# Increment this version number whenever the schema of discovery.yml is changed.
# See https://www.meltano.com/docs/contributor-guide.html#discovery-yml-version for more information.
VERSION = 12


class DiscoveryFile(Canonical):
    def __init__(self, version=1, **plugins):
        super().__init__(version=int(version))

        for plugin_type in PluginType:
            self[plugin_type] = []

        for plugin_type, plugin_defs in plugins.items():
            for plugin_def in plugin_defs:
                plugin = Plugin(
                    plugin_type,
                    plugin_def.pop("name"),
                    plugin_def.pop("namespace"),
                    **plugin_def,
                )
                self[plugin_type].append(plugin)

    @classmethod
    def version(cls, attrs):
        return int(attrs.get("version", 1))


class PluginDiscoveryService(Versioned):
    __version__ = VERSION

    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        discovery: Optional[Dict] = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)

        self._discovery_version = None
        self._discovery = None
        if discovery:
            self._discovery_version = DiscoveryFile.version(discovery)
            self._discovery = DiscoveryFile.parse(discovery)

    @property
    def file_version(self):
        return self._discovery_version

    @property
    def discovery_url(self):
        discovery_url = ProjectSettingsService(self.project).get("discovery_url")

        if not discovery_url or not re.match(r"^https?://", discovery_url):
            return None

        return discovery_url

    @property
    def discovery(self):
        """
        Return first compatible discovery manifest from these locations:

        - project local `discovery.yml`
        - `discovery_url` project setting
        - .meltano/cache/discovery.yml
        - meltano.core.bundle
        """
        if self._discovery:
            return self._discovery

        loaders = [
            (
                self.load_local_discovery,
                "your project's local `discovery.yml` manifest",
            ),
            (
                self.load_remote_discovery,
                f"the `discovery.yml` manifest received from {self.discovery_url}",
            ),
            (self.load_cached_discovery, "the cached `discovery.yml` manifest"),
            (self.load_bundled_discovery, "the bundled `discovery.yml` manifest"),
        ]

        errored = False
        for loader, description in loaders:
            if errored:
                logging.warning(f"Falling back on {description}...")

            try:
                loader()
            except IncompatibleVersionError as err:
                errored = True

                logging.warning(
                    f"{description.capitalize()} has version {err.file_version}, while this version of Meltano requires version {err.version}."
                )
                if err.file_version > err.version:
                    logging.warning(
                        "Please install the latest compatible version of Meltano using `meltano upgrade`."
                    )
            except DiscoveryInvalidError as err:
                errored = True
                logging.error(f"{description.capitalize()} could not be parsed.")
                logging.debug(str(err))

            if self._discovery:
                return self._discovery

        raise DiscoveryInvalidError("No valid `discovery.yml` manifest could be found")

    def load_local_discovery(self):
        try:
            with self.project.root_dir("discovery.yml").open() as local_discovery:
                return self.load_discovery(local_discovery)
        except FileNotFoundError:
            pass

    def load_remote_discovery(self):
        if not self.discovery_url:
            return

        try:
            response = requests.get(self.discovery_url)
            response.raise_for_status()

            remote_discovery = io.StringIO(response.text)
            discovery = self.load_discovery(remote_discovery, cache=True)

            return discovery
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as err:
            logging.debug("Remote `discovery.yml` manifest could not be downloaded.")
            logging.debug(str(err))
            pass

    def load_cached_discovery(self):
        try:
            with self.cached_discovery_file.open() as cached_discovery:
                return self.load_discovery(cached_discovery)
        except FileNotFoundError:
            pass

    def load_bundled_discovery(self):
        with bundle.find("discovery.yml").open() as bundled_discovery:
            discovery = self.load_discovery(bundled_discovery, cache=True)

        return discovery

    def load_discovery(self, discovery_file, cache=False):
        try:
            discovery_yaml = yaml.safe_load(discovery_file)

            self._discovery_version = DiscoveryFile.version(discovery_yaml)
            self.ensure_compatible()

            self._discovery = DiscoveryFile.parse(discovery_yaml)

            if cache:
                self.cache_discovery()

            return self._discovery
        except IncompatibleVersionError:
            raise
        except (yaml.YAMLError, Exception) as err:
            raise DiscoveryInvalidError(str(err))

    def cache_discovery(self):
        with self.cached_discovery_file.open("w") as cached_discovery:
            yaml.dump(
                self._discovery,
                cached_discovery,
                default_flow_style=False,
                sort_keys=False,
            )

    @property
    def cached_discovery_file(self):
        return self.project.meltano_dir("cache", "discovery.yml")

    def plugins(self) -> Iterator[Plugin]:
        discovery_plugins = (
            plugin
            for plugin_type in PluginType
            for plugin in self.discovery[plugin_type]
        )

        yield from self.custom_plugins()
        yield from discovery_plugins

    def custom_plugins(self) -> Iterator[Plugin]:
        # some plugins in the Meltano file might be custom, thus they
        # serve both as `PluginInstall` and `Plugin`
        custom_plugin_type_defs = (
            (custom_plugin.type, custom_plugin.canonical())
            for custom_plugin in self.config_service.plugins()
            if custom_plugin.is_custom()
        )

        custom_plugins = (
            Plugin(
                plugin_type,
                custom_plugin_def.pop("name"),
                custom_plugin_def.pop("namespace"),
                **custom_plugin_def,
            )
            for plugin_type, custom_plugin_def in custom_plugin_type_defs
        )

        yield from custom_plugins

    def find_plugin(self, plugin_type: PluginType, plugin_name: str):
        name, _ = PluginRef.parse_name(plugin_name)
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (plugin.type == plugin_type and plugin.name == name)
            )
        except StopIteration:
            raise PluginNotFoundError(name)

    def find_plugin_by_namespace(self, plugin_type: PluginType, namespace: str):
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (plugin.type == plugin_type and plugin.namespace == namespace)
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def discover(self, plugin_type: PluginType = None):
        """Return a pretty printed list of available plugins."""
        enabled_plugin_types = (plugin_type,) if plugin_type else iter(PluginType)

        return {
            plugin_type: [p.name for p in plugins]
            for plugin_type, plugins in groupby(self.plugins(), lambda p: p.type)
            if plugin_type in enabled_plugin_types
        }

    def list_discovery(self, discovery):
        return "\n".join(
            plugin.name for plugin in self.plugins() if plugin.type == discovery
        )



import os
import json
import yaml
import logging
from typing import List

from .project import Project
from .plugin import PluginType, PluginInstall, PluginRef
from .plugin_discovery_service import PluginDiscoveryService
from .plugin.factory import plugin_factory
from .config_service import ConfigService, PluginAlreadyAddedException


class PluginNotSupportedException(Exception):
    pass


class MissingPluginException(Exception):
    pass


class ProjectAddService:
    def __init__(
        self,
        project: Project,
        plugin_discovery_service: PluginDiscoveryService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str, **kwargs) -> PluginInstall:
        plugin = self.discovery_service.find_plugin(plugin_type, plugin_name)
        installed = plugin.as_installed()
        return self.config_service.add_to_file(installed)

    def add_related(
        self,
        target_plugin: PluginInstall,
        plugin_types: List[PluginType] = list(PluginType),
    ):
        try:
            plugin_types.remove(target_plugin.type)
        except ValueError:
            pass

        related_plugins = [
            plugin
            for plugin in self.discovery_service.plugins()
            if plugin.namespace == target_plugin.namespace
            and plugin.type in plugin_types
        ]

        added_plugins = []
        for plugin in related_plugins:
            try:
                plugin_install = self.add(plugin.type, plugin.name)

                added_plugins.append(plugin_install)
            except PluginAlreadyAddedException:
                pass

        return added_plugins



import os
import click
import subprocess
import psutil
import meltano
import logging
from pathlib import Path
from sqlalchemy import create_engine
from click_default_group import DefaultGroup

from meltano.core.project import Project
from meltano.core.meltano_invoker import MeltanoInvoker
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationService
from meltano.core.upgrade_service import UpgradeService, UpgradeError
from . import cli
from .params import project


@cli.group(cls=DefaultGroup, default="all", default_if_no_args=True)
@project()
@click.pass_context
def upgrade(ctx, project):
    engine, _ = project_engine(project)
    upgrade_service = UpgradeService(engine, project)
    ctx.obj["upgrade_service"] = upgrade_service


@upgrade.command()
@click.option("--pip_url", type=str, envvar="MELTANO_UPGRADE_PIP_URL")
@click.option("--force", is_flag=True, default=False, envvar="MELTANO_UPGRADE_FORCE")
@click.option("--skip-package", is_flag=True, default=False)
@click.pass_context
def all(ctx, pip_url, force, skip_package):
    project = ctx.obj["project"]
    upgrade_service = ctx.obj["upgrade_service"]

    try:
        if skip_package:
            upgrade_service.update_files()

            click.echo()
            upgrade_service.migrate_database()

            click.echo()
            upgrade_service.compile_models()

            if not os.getenv("MELTANO_PACKAGE_UPGRADED", False):
                click.echo()
                click.secho("Your Meltano project has been upgraded!", fg="green")
        else:
            package_upgraded = upgrade_service.upgrade_package(
                pip_url=pip_url, force=force
            )
            if package_upgraded:
                # Shell out instead of calling `upgrade_service` methods to
                # ensure the latest code is used.
                click.echo()
                run = MeltanoInvoker(project).invoke(
                    ["upgrade", "--skip-package"],
                    env={"MELTANO_PACKAGE_UPGRADED": "true"},
                )

                if run.returncode == 0:
                    click.echo()
                    click.secho(
                        "Meltano and your Meltano project have been upgraded!",
                        fg="green",
                    )
            else:
                click.echo(
                    "Then, run `meltano upgrade --skip-package` to upgrade your project based on the latest version."
                )
    except UpgradeError as up:
        click.secho(str(up), fg="red")
        raise click.Abort()


@upgrade.command()
@click.option("--pip_url", type=str, envvar="MELTANO_UPGRADE_PIP_URL")
@click.option("--force", is_flag=True, default=False, envvar="MELTANO_UPGRADE_FORCE")
@click.pass_context
def package(ctx, **kwargs):
    try:
        ctx.obj["upgrade_service"].upgrade_package(**kwargs)
    except UpgradeError as up:
        click.secho(str(up), fg="red")
        raise click.Abort()


@upgrade.command()
@click.pass_context
def files(ctx):
    try:
        ctx.obj["upgrade_service"].update_files()
    except UpgradeError as up:
        click.secho(str(up), fg="red")
        raise click.Abort()


@upgrade.command()
@click.pass_context
def database(ctx):
    try:
        ctx.obj["upgrade_service"].migrate_database()
    except UpgradeError as up:
        click.secho(str(up), fg="red")
        raise click.Abort()


@upgrade.command()
@click.pass_context
def models(ctx):
    try:
        ctx.obj["upgrade_service"].compile_models()
    except UpgradeError as up:
        click.secho(str(up), fg="red")
        raise click.Abort()



import os
import sqlalchemy
import logging
import re
import dotenv
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Iterable, Dict, List
from enum import Enum

from meltano.core.utils import (
    find_named,
    setting_env,
    NotFound,
    flatten,
    set_at_path,
    pop_at_path,
)
from .setting_definition import SettingDefinition
from .setting import Setting
from .config_service import ConfigService
from .error import Error


class SettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, name: str):
        super().__init__(f"Cannot find setting {name}")


class SettingValueSource(str, Enum):
    CONFIG_OVERRIDE = "config_override"  # 0
    ENV = "env"  # 1
    DOTENV = "dotenv"  # 2
    MELTANO_YML = "meltano_yml"  # 3
    DB = "db"  # 4
    DEFAULT = "default"  # 5

    @property
    def label(self):
        labels = {
            self.CONFIG_OVERRIDE: "a command line flag",
            self.ENV: "the environment",
            self.DOTENV: "`.env`",
            self.MELTANO_YML: "`meltano.yml`",
            self.DB: "the system database",
            self.DEFAULT: "the default",
        }
        return labels[self]


class SettingValueStore(str, Enum):
    DOTENV = "dotenv"
    MELTANO_YML = "meltano_yml"
    DB = "db"


# sentinel value to use to prevent leaking sensitive data
REDACTED_VALUE = "(redacted)"


class SettingsService(ABC):
    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        show_hidden=True,
        env_override={},
        config_override={},
    ):
        self.project = project

        self.config_service = config_service or ConfigService(project)

        self.show_hidden = show_hidden

        self.env_override = env_override
        self.config_override = config_override

    @property
    @abstractmethod
    def _env_namespace(self) -> str:
        pass

    @property
    @abstractmethod
    def _db_namespace(self) -> str:
        pass

    @property
    @abstractmethod
    def _definitions(self) -> List[SettingDefinition]:
        pass

    @property
    @abstractmethod
    def _meltano_yml_config(self) -> Dict:
        pass

    @abstractmethod
    def _update_meltano_yml_config(self):
        pass

    @property
    def flat_meltano_yml_config(self):
        return flatten(self._meltano_yml_config, "dot")

    @property
    def env(self):
        return {**os.environ, **self.env_override}

    @classmethod
    def is_kind_redacted(cls, kind) -> bool:
        return kind in ("password", "oauth")

    @classmethod
    def unredact(cls, values: dict) -> Dict:
        """
        Removes any redacted values in a dictionary.
        """

        return {k: v for k, v in values.items() if v != REDACTED_VALUE}

    def config_with_metadata(self, sources: List[SettingValueSource] = None, **kwargs):
        config = {}
        for setting in self.definitions():
            value, source = self.get_with_source(setting.name, **kwargs)
            if sources and source not in sources:
                continue

            config[setting.name] = {
                "value": value,
                "source": source,
                "setting": setting,
            }

        return config

    def as_dict(self, *args, **kwargs) -> Dict:
        full_config = self.config_with_metadata(*args, **kwargs)

        return {key: config["value"] for key, config in full_config.items()}

    def as_env(self, *args, **kwargs) -> Dict[str, str]:
        full_config = self.config_with_metadata(*args, **kwargs)

        return {
            self.setting_env(config["setting"]): str(config["value"])
            for key, config in full_config.items()
            if config["value"] is not None
        }

    def get_with_source(self, name: str, redacted=False, session=None):
        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        def config_override_getter():
            try:
                return self.config_override[name]
            except KeyError:
                return None

        def env_getter(env=self.env):
            if not setting_def:
                return None

            env_key = self.setting_env(setting_def)
            env_getters = {
                env_key: lambda env: env[env_key],
                **setting_def.env_alias_getters,
            }

            for key, getter in env_getters.items():
                try:
                    return getter(env)
                except KeyError:
                    pass
                else:
                    logging.debug(
                        f"Found ENV variable {key} for {self._env_namespace}:{name}"
                    )

            return None

        def dotenv_getter():
            return env_getter(env=dotenv.dotenv_values(self.project.dotenv))

        def meltano_yml_getter():
            try:
                value = self.flat_meltano_yml_config[name]
                return self.expand_env_vars(value)
            except KeyError:
                return None

        def db_getter():
            if not session:
                return None

            try:
                return (
                    session.query(Setting)
                    .filter_by(namespace=self._db_namespace, name=name, enabled=True)
                    .one()
                    .value
                )
            except sqlalchemy.orm.exc.NoResultFound:
                return None

        def default_getter():
            if not setting_def:
                return None
            return self.expand_env_vars(setting_def.value)

        config_getters = {
            SettingValueSource.CONFIG_OVERRIDE: config_override_getter,
            SettingValueSource.ENV: env_getter,
            SettingValueSource.DOTENV: dotenv_getter,
            SettingValueSource.MELTANO_YML: meltano_yml_getter,
            SettingValueSource.DB: db_getter,
            SettingValueSource.DEFAULT: default_getter,
        }

        for source, getter in config_getters.items():
            value = getter()
            if value is not None:
                break

        if setting_def:
            value = setting_def.cast_value(value)

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and self.is_kind_redacted(setting_def.kind):
                value = REDACTED_VALUE

        return value, source

    def get(self, *args, **kwargs):
        value, _ = self.get_with_source(*args, **kwargs)
        return value

    def set(
        self, path: List[str], value, store=SettingValueStore.MELTANO_YML, session=None
    ):
        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        if value == REDACTED_VALUE:
            return

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        if setting_def:
            value = setting_def.cast_value(value)

        def dotenv_setter():
            if not setting_def:
                return None

            env_key = self.setting_env(setting_def)

            dotenv_file = self.project.dotenv

            if value is None:
                if dotenv_file.exists():
                    for key in [env_key, *setting_def.env_alias_getters.keys()]:
                        dotenv.unset_key(dotenv_file, key)
            else:
                if dotenv_file.exists():
                    for key in setting_def.env_alias_getters.keys():
                        dotenv.unset_key(dotenv_file, key)
                else:
                    dotenv_file.touch()

                dotenv.set_key(dotenv_file, env_key, str(value))

            return True

        def meltano_yml_setter():
            config = self._meltano_yml_config

            if value is None:
                config.pop(name, None)
                pop_at_path(config, name, None)
                pop_at_path(config, path, None)
            else:
                if len(path) > 1:
                    config.pop(name, None)

                if name.split(".") != path:
                    pop_at_path(config, name, None)

                set_at_path(config, path, value)

            self._update_meltano_yml_config()
            return True

        def db_setter():
            if not session:
                return None

            if value is None:
                session.query(Setting).filter_by(
                    namespace=self._db_namespace, name=name
                ).delete()
            else:
                setting = Setting(
                    namespace=self._db_namespace, name=name, value=value, enabled=True
                )
                session.merge(setting)

            session.commit()
            return True

        config_setters = {
            SettingValueStore.DOTENV: dotenv_setter,
            SettingValueStore.MELTANO_YML: meltano_yml_setter,
            SettingValueStore.DB: db_setter,
        }

        if not config_setters[store]():
            return

        return value

    def unset(self, path: List[str], **kwargs):
        return self.set(path, None, **kwargs)

    def reset(self, store=SettingValueStore.MELTANO_YML, session=None):
        def dotenv_resetter():
            dotenv_file = self.project.dotenv

            if dotenv_file.exists():
                dotenv_file.unlink()

            return True

        def meltano_yml_resetter():
            self._meltano_yml_config.clear()
            self._update_meltano_yml_config()
            return True

        def db_resetter():
            if not session:
                return None

            session.query(Setting).filter_by(namespace=self._db_namespace).delete()
            session.commit()
            return True

        config_resetters = {
            SettingValueStore.DOTENV: dotenv_resetter,
            SettingValueStore.MELTANO_YML: meltano_yml_resetter,
            SettingValueStore.DB: db_resetter,
        }

        if not config_resetters[store]():
            return False

        return True

    def definitions(self) -> Iterable[Dict]:
        definitions = deepcopy(self._definitions)
        definition_names = set(s.name for s in definitions)

        definitions.extend(
            (
                SettingDefinition.from_key_value(k, v)
                for k, v in self.flat_meltano_yml_config.items()
                if k not in definition_names
            )
        )

        settings = []
        for setting in definitions:
            if setting.kind == "hidden" and not self.show_hidden:
                continue

            settings.append(setting)

        return settings

    def find_setting(self, name: str) -> SettingDefinition:
        try:
            return find_named(self.definitions(), name)
        except NotFound as err:
            raise SettingMissingError(name) from err

    def setting_env(self, setting_def):
        return setting_def.env or setting_env(self._env_namespace, setting_def.name)

    def expand_env_vars(self, raw_value):
        if not isinstance(raw_value, str):
            return raw_value

        # find viable substitutions
        var_matcher = re.compile(
            """
            \$                 # starts with a '$'
            (?:                # either $VAR or ${VAR}
                {(\w+)}|(\w+)  # capture the variable name as group[0] or group[1]
            )
            """,
            re.VERBOSE,
        )

        def subst(match) -> str:
            try:
                # the variable can be in either group
                var = next(var for var in match.groups() if var)
                val = str(self.env[var])

                if not val:
                    logging.warning(f"Variable {var} is empty.")

                return val
            except KeyError as e:
                logging.warning(f"Variable {var} is missing from the environment.")
                return None

        fullmatch = re.fullmatch(var_matcher, raw_value)
        if fullmatch:
            # If the entire value is an env var reference, return None if it isn't set
            return subst(fullmatch)

        return re.sub(var_matcher, subst, raw_value)



from typing import List

from meltano.core.utils import truthy
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq


class SettingDefinition(NameEq, Canonical):
    def __init__(
        self,
        name: str = None,
        env: str = None,
        env_aliases: List[str] = [],
        kind: str = None,
        value=None,
        label: str = None,
        documentation: str = None,
        description: str = None,
        tooltip: str = None,
        options: list = [],
        oauth: dict = {},
        placeholder: str = None,
        protected: bool = None,
        custom: bool = False,
        **attrs,
    ):
        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            name=name,
            env=env,
            env_aliases=env_aliases,
            kind=kind,
            value=value,
            label=label,
            documentation=documentation,
            description=description,
            tooltip=tooltip,
            options=options,
            oauth=oauth,
            placeholder=placeholder,
            protected=protected,
            _custom=custom,
            **attrs,
        )

        self._verbatim.add("value")

    @classmethod
    def from_key_value(cls, key, value):
        kind = None
        if isinstance(value, bool):
            kind = "boolean"

        return cls(name=key, kind=kind, custom=True)

    @property
    def env_alias_getters(self):
        getters = {}

        for alias in self.env_aliases:
            key = alias

            if alias.startswith("!") and self.kind == "boolean":
                key = alias[1:]
                getter = lambda env, key=key: str(not truthy(env[key]))
            else:
                getter = lambda env, key=alias: env[key]

            getters[key] = getter

        return getters

    def cast_value(self, value):
        if isinstance(value, str):
            if self.kind == "boolean":
                return truthy(value)
            elif self.kind == "integer":
                return int(value)

        return value



from typing import Iterable, Dict, List

from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.settings_service import (
    SettingsService,
    SettingMissingError,
    SettingValueSource,
    SettingValueStore,
    REDACTED_VALUE,
)
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin import PluginRef, PluginType, Plugin, PluginInstall, Profile
from meltano.core.plugin.error import PluginMissingError


class PluginSettingsService(SettingsService):
    def __init__(
        self,
        project: Project,
        plugin: PluginRef,
        *args,
        plugin_discovery_service: PluginDiscoveryService = None,
        **kwargs,
    ):
        super().__init__(project, *args, **kwargs)

        self.plugin = plugin

        discovery_service = plugin_discovery_service or PluginDiscoveryService(
            self.project, config_service=self.config_service
        )
        self.plugin_def = discovery_service.find_plugin(
            self.plugin.type, self.plugin.name
        )
        self._plugin_install = None

        project_settings_service = ProjectSettingsService(
            self.project, config_service=self.config_service
        )

        self.env_override = {
            **project_settings_service.env,
            **project_settings_service.as_env(),
            **self.env_override,
        }

    @property
    def plugin_install(self):
        if self._plugin_install is None:
            self._plugin_install = self.config_service.get_plugin(self.plugin)

        return self._plugin_install

    @property
    def _env_namespace(self):
        return self.plugin_def.namespace

    @property
    def _db_namespace(self):
        return self.plugin.qualified_name

    @property
    def _definitions(self):
        return self.plugin_def.settings

    @property
    def _meltano_yml_config(self):
        try:
            return self.plugin_install.current_config
        except PluginMissingError:
            return {}

    def _update_meltano_yml_config(self):
        self.config_service.update_plugin(self.plugin_install)

    def profile_with_config(self, profile: Profile, **kwargs):
        self.plugin_install.use_profile(profile)

        full_config = self.config_with_metadata(**kwargs)

        return {
            **profile.canonical(),
            "config": {key: config["value"] for key, config in full_config.items()},
            "config_sources": {
                key: config["source"] for key, config in full_config.items()
            },
        }

    def profiles_with_config(self, **kwargs) -> List[Dict]:
        return [
            self.profile_with_config(profile, **kwargs)
            for profile in (Profile.DEFAULT, *self.plugin_install.profiles)
        ]

    def set(self, *args, store=SettingValueStore.DB, **kwargs):
        return super().set(*args, **kwargs, store=store)

    def unset(self, *args, store=SettingValueStore.DB, **kwargs):
        return super().unset(*args, **kwargs, store=store)

    def reset(self, *args, store=SettingValueStore.DB, **kwargs):
        return super().reset(*args, **kwargs, store=store)



import json
import logging
import subprocess
from typing import Dict
from jsonschema import Draft4Validator

from meltano.core.utils import file_has_data, truthy
from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError

from . import SingerPlugin, PluginType
from .catalog import (
    MetadataExecutor,
    MetadataRule,
    property_breadcrumb,
    select_metadata_rules,
)


def config_metadata_rules(config):
    rules = []
    for key, value in config.items():
        if not key.startswith("metadata."):
            continue

        # metadata.<tap_stream_id>.<key>
        # metadata.<tap_stream_id>.<prop>.<key>
        # metadata.<tap_stream_id>.<prop>.<subprop>.<key>
        # metadata.<tap_stream_id>.properties.<prop>.<key>
        # metadata.<tap_stream_id>.properties.<prop>.properties.<subprop>.<key>
        _, tap_stream_id, *props, key = key.split(".")

        if key == "selected":
            value = truthy(value)

        rules.append(
            MetadataRule(
                tap_stream_id=tap_stream_id,
                breadcrumb=property_breadcrumb(props),
                key=key,
                value=value,
            )
        )

    return rules


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    def exec_args(self, plugin_invoker):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            if "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]

        state_path = plugin_invoker.files["state"]
        if "state" in plugin_invoker.capabilities and file_has_data(state_path):
            args += ["--state", state_path]

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke", can_fail=True)
    def run_discovery_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args:
            return

        try:
            self.run_discovery(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            return

    def run_discovery(self, plugin_invoker, exec_args=[]):
        if not "discover" in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support schema discovery"
            )

        properties_file = plugin_invoker.files["catalog"]

        with properties_file.open("w") as catalog:
            result = plugin_invoker.invoke(
                "--discover",
                stdout=catalog,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            stdout, stderr = result.communicate()
            exit_code = result.returncode

        if exit_code != 0:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Schema discovery failed: command {plugin_invoker.exec_args('--discover')} returned {exit_code}: {stderr.rstrip()}"
            )

        # test for the schema to be a valid catalog
        try:
            with properties_file.open("r") as catalog:
                schema_valid = Draft4Validator.check_schema(json.load(catalog))
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                "Schema discovery failed: invalid catalog output by --discovery."
            ) from err

    @hook("before_invoke", can_fail=True)
    def apply_metadata_rules_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args:
            return

        try:
            self.apply_metadata_rules(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            return

    def apply_metadata_rules(self, plugin_invoker, exec_args=[]):
        if (
            not "catalog" in plugin_invoker.capabilities
            and not "properties" in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support entity selection or metadata rules"
            )

        properties_file = plugin_invoker.files["catalog"]

        try:
            with properties_file.open() as catalog:
                schema = json.load(catalog)

            metadata_rules = [
                *select_metadata_rules(["!*.*"]),
                *select_metadata_rules(plugin_invoker.select),
                *config_metadata_rules(plugin_invoker.plugin_config),
            ]

            metadata_executor = MetadataExecutor(metadata_rules)
            metadata_executor.visit(schema)

            with properties_file.open("w") as catalog:
                json.dump(schema, catalog)
        except FileNotFoundError as err:
            raise PluginExecutionError(
                f"Applying metadata rules failed: catalog file is missing."
            ) from err
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Applying metadata rules failed: catalog file is invalid: {properties_file}"
            ) from err



import click
import json

from . import cli
from .params import project

from meltano.core.db import project_engine
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginMissingError
from meltano.core.config_service import ConfigService
from meltano.core.plugin.settings_service import (
    PluginSettingsService,
    SettingValueSource,
    SettingValueStore,
)


@cli.group(invoke_without_command=True)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.argument("plugin_name")
@click.option("--format", type=click.Choice(["json", "env"]), default="json")
@project(migrate=True)
@click.pass_context
def config(ctx, project, plugin_type, plugin_name, format):
    try:
        plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None

        config = ConfigService(project)
        plugin = config.find_plugin(
            plugin_name, plugin_type=plugin_type, configurable=True
        )
    except PluginMissingError:
        if plugin_name == "meltano":
            plugin = None
        else:
            raise

    _, Session = project_engine(project)
    session = Session()
    try:
        if plugin:
            settings = PluginSettingsService(project, plugin)
        else:
            settings = ProjectSettingsService(project)

        ctx.obj["settings"] = settings
        ctx.obj["session"] = session

        if ctx.invoked_subcommand is None:
            if format == "json":
                config = settings.as_dict(session=session)
                print(json.dumps(config))
            elif format == "env":
                for env, value in settings.as_env(session=session).items():
                    print(f"{env}={value}")
    finally:
        session.close()


@config.command()
@click.argument("setting_name", nargs=-1, required=True)
@click.argument("value")
@click.option(
    "--store",
    type=click.Choice(list(SettingValueStore)),
    default=SettingValueStore.MELTANO_YML,
)
@click.pass_context
def set(ctx, setting_name, value, store):
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    path = list(setting_name)
    settings.set(path, value, store=store, session=session)


@config.command()
@click.argument("setting_name", nargs=-1, required=True)
@click.option(
    "--store",
    type=click.Choice(list(SettingValueStore)),
    default=SettingValueStore.MELTANO_YML,
)
@click.pass_context
def unset(ctx, setting_name, store):
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    path = list(setting_name)
    settings.unset(path, store=store, session=session)


@config.command()
@click.option(
    "--store",
    type=click.Choice(list(SettingValueStore)),
    default=SettingValueStore.MELTANO_YML,
)
@click.pass_context
def reset(ctx, store):
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    settings.reset(store=store, session=session)


@config.command("list")
@click.pass_context
def list_settings(ctx):
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    full_config = settings.config_with_metadata(session=session)
    for name, config_metadata in full_config.items():
        value = config_metadata["value"]
        source = config_metadata["source"]
        setting_def = config_metadata["setting"]

        if setting_def._custom:
            click.echo("custom: ", nl=False)

        click.secho(name, fg="blue", nl=False)

        env_keys = [settings.setting_env(setting_def), *setting_def.env_aliases]
        click.echo(f" [env: {', '.join(env_keys)}]", nl=False)

        current_value = click.style(f"{value!r}", fg="green")
        if source is SettingValueSource.DEFAULT:
            click.echo(f" current value: {current_value}", nl=False)

            # The default value and the current value may not match
            # if env vars have been expanded
            if setting_def.value == value:
                click.echo(" (from default)")
            else:
                click.echo(f" (from default: {setting_def.value!r})")
        else:
            if setting_def.value is not None:
                click.echo(f" (default: {setting_def.value!r})", nl=False)

            click.echo(f" current value: {current_value} (from {source.label})")

        if setting_def.description:
            click.echo("\t", nl=False)
            if setting_def.label:
                click.echo(f"{setting_def.label}: ", nl=False)
            click.echo(f"{setting_def.description}")



import yaml
import fnmatch
import copy
from typing import Dict, Union
from collections import namedtuple
from enum import Enum
from typing import Optional, Iterable

from meltano.core.setting_definition import SettingDefinition
from meltano.core.behavior.hookable import HookObject
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq
from meltano.core.utils import compact, find_named, NotFound


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class Profile(NameEq, Canonical):
    def __init__(self, name: str = None, label: str = None, config={}):
        super().__init__(name=name, label=label, config=config)


Profile.DEFAULT = Profile(name="default", label="Default")


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    TRANSFORMS = "transforms"
    MODELS = "models"
    DASHBOARDS = "dashboards"
    ORCHESTRATORS = "orchestrators"
    TRANSFORMERS = "transformers"
    FILES = "files"

    def __str__(self):
        return self.value

    @property
    def descriptor(self):
        if self is self.__class__.FILES:
            return "file bundle"

        return self.singular

    @property
    def singular(self):
        """Makes it singular for `meltano add PLUGIN_TYPE`"""
        return self.value[:-1]

    @property
    def verb(self):
        return self.value[:-3]

    @classmethod
    def value_exists(cls, value):
        return value in cls._value2member_map_

    @classmethod
    def cli_arguments(cls):
        args = [type.singular for type in cls]
        args.extend([type for type in cls])
        return args

    @classmethod
    def from_cli_argument(cls, value):
        if not value.endswith("s"):
            value += "s"

        return cls(value)


class PluginRef:
    def __init__(self, plugin_type: Union[str, PluginType], name: str):
        self._type = (
            plugin_type
            if isinstance(plugin_type, PluginType)
            else PluginType(plugin_type)
        )
        self.name, self._current_profile_name = self.parse_name(name)

    @classmethod
    def parse_name(cls, name: str):
        name, *profile_name = name.split("@")
        profile_name = next(iter(profile_name), Profile.DEFAULT.name)

        return (name, profile_name)

    @property
    def type(self):
        return self._type

    @property
    def current_profile_name(self):
        return self._current_profile_name

    @property
    def full_name(self):
        return f"{self.name}@{self.current_profile_name}"

    @property
    def qualified_name(self):
        parts = (self.type, self.name, self.current_profile_name)

        return ".".join(compact(parts))

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __hash__(self):
        return hash((self.type, self.name, self.current_profile_name))


class PluginInstall(HookObject, Canonical, PluginRef):
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        label: Optional[str] = None,
        logo_url: Optional[str] = None,
        namespace: Optional[str] = None,
        pip_url: Optional[str] = None,
        executable: str = None,
        capabilities: list = [],
        settings: list = [],
        select: list = [],
        config={},
        profiles: list = [],
        **attrs,
    ):
        super().__init__(
            plugin_type,
            name,
            # Attributes will be listed in meltano.yml in this order:
            label=label,
            logo_url=logo_url,
            namespace=namespace,
            pip_url=pip_url,
            executable=executable,
            capabilities=list(capabilities),
            settings=list(map(SettingDefinition.parse, settings)),
            select=list(select),
            config=copy.deepcopy(config),
            profiles=list(map(Profile.parse, profiles)),
            **attrs,
        )

        self._extras = attrs

    def is_installable(self):
        return self.pip_url is not None

    def is_invokable(self):
        return self.is_installable()

    def is_configurable(self):
        return True

    def should_add_to_file(self, project):
        return True

    def is_custom(self):
        return self.namespace is not None

    def get_profile(self, profile_name: str) -> Profile:
        if profile_name == Profile.DEFAULT.name:
            return Profile.DEFAULT

        return find_named(self.profiles, profile_name)

    def use_profile(self, profile_or_name: Union[str, Profile]):
        if profile_or_name is None:
            profile = Profile.DEFAULT
        elif isinstance(profile_or_name, Profile):
            profile = self.get_profile(profile_or_name.name)
        else:
            profile = self.get_profile(profile_or_name)

        self._current_profile_name = profile.name

    @property
    def current_profile(self):
        return self.get_profile(self.current_profile_name)

    @property
    def current_config(self):
        if self.current_profile is Profile.DEFAULT:
            return self.config

        return self.current_profile.config

    def exec_args(self, files: Dict):
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return dict()

    @property
    def output_files(self):
        return dict()

    def add_select_filter(self, filter: str):
        self.select.append(filter)

    def add_profile(self, name: str, config: dict = None, label: str = None):
        profile = Profile(name=name, config=config, label=label)

        self.profiles.append(profile)
        return profile


class Plugin(Canonical, PluginRef):
    """
    Args:
    name: The unique name for the installed plugin
    pip_url: The pip-compatible installation URI, like `git+https://…` or `-e /path/to/pkg`
    executable: The plugin executable name (default: <name>)
    """

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: str,
        label: Optional[str] = None,
        logo_url: Optional[str] = None,
        description=None,
        docs=None,
        pip_url: Optional[str] = None,
        capabilities: list = [],
        settings_group_validation: list = [],
        settings: list = [],
        select: list = [],
        **attrs,
    ):
        super().__init__(
            plugin_type,
            name,
            # Attributes will be listed in meltano.yml in this order:
            label=label,
            logo_url=logo_url,
            description=description,
            docs=docs,
            namespace=namespace,
            pip_url=pip_url,
            capabilities=list(capabilities),
            settings_group_validation=list(settings_group_validation),
            settings=list(map(SettingDefinition.parse, settings)),
            select=list(select),
            **attrs,
        )

        self._extras = attrs

    def as_installed(self, custom=False) -> PluginInstall:
        extras = self._extras.copy()
        if custom:
            extras.update(
                {
                    "namespace": self.namespace,
                    "capabilities": self.capabilities,
                    "settings": self.settings,
                }
            )

        return PluginInstall(
            self.type, self.name, pip_url=self.pip_url, select=self.select, **extras
        )
